#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

static int fd;
static unsigned long old_state, new_state;
static int use_nonblock;
static int use_async;
static int use_poll;

/*
 * When the -a option is given, this handler will be called whenever
 * the pin state changes.
 */
static void sigio_handler(int sig, siginfo_t *info, void *context)
{
	struct pollfd fds = {
		.fd = fd,
		.events = POLLIN,
	};
	unsigned long pin_state;
	int ret;

	if (use_poll) {
		/* There should be data available immediately */
		ret = poll(&fds, 1, 0);
		if (ret < 0) {
			perror("sigio_handler: poll:");
			return;
		} else if (ret == 0) {
			fprintf(stderr,
				"No data available in sigio_handler!\n");
			return;
		} else {
			fprintf(stderr, "sigio_handler: poll() returned %d\n",
				ret);
		}
	}

	/*
	 * Even if use_nonblock is set, this should never fail as we
	 * shouldn't be here if no new data is available.
	 */
	if (read(fd, &pin_state, sizeof(pin_state)) < 0)
		perror("read error in sigio_handler");
	else
		new_state = pin_state;
}

static void display_usage(const char *cmd)
{
	fprintf(stderr, "Usage: %s [-ahnp] [gpio-device]\n\n", cmd);
	fprintf(stderr, "  -a     Use asynchronous I/O (O_ASYNC)\n");
	fprintf(stderr, "  -h     Display usage information (this message)\n");
	fprintf(stderr, "  -n     Use nonblocking I/O (O_NONBLOCK)\n");
	fprintf(stderr, "  -p     Use poll() to wait for I/O\n");

	fprintf(stderr,
		"\nIf no gpio-device is specified, /dev/gpio0 is used\n");
}

int main(int argc, char **argv)
{
	const char *dev_name = "/dev/gpio0";

	while (1) {
		int c;

		c = getopt(argc, argv, "ahnp");
		if (c == -1)
			break;

		switch (c) {
		case 'a':
			use_async = 1;
			break;
		case 'n':
			use_nonblock = 1;
			break;
		case 'p':
			use_poll = 1;
			break;
		case 'h':
		case '?':
			display_usage(argv[0]);
			return 1;
		default:
			abort();
		}
	}

	if (optind < argc)
		dev_name = argv[optind++];

	fd = open(dev_name, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "failed to open %s: %s\n",
			dev_name, strerror(errno));
		return 1;
	}

	if (use_async) {
		struct sigaction sigio_action = {
			.sa_sigaction = sigio_handler,
			.sa_flags = SA_SIGINFO,
		};
		int flags;

		/*
		 * Don't mask any signals when handling SIGIO (except
		 * SIGIO itself, which is masked by default.)
		 */
		sigemptyset(&sigio_action.sa_mask);
		if (sigaction(SIGIO, &sigio_action, NULL)) {
			perror("failed to install SIGIO handler");
			return 1;
		}

		/* Tell the kernel that our process will take care of SIGIO */
		if (fcntl(fd, F_SETOWN, getpid())) {
			perror("failed to set fd owner");
			return 1;
		}

		/* Enable asynchronous I/O */
		flags = fcntl(fd, F_GETFL);
		if (fcntl(fd, F_SETFL, flags | O_ASYNC) < 0) {
			perror("failed to set O_ASYNC");
			return 1;
		}
	}
	if (use_nonblock) {
		int flags;

		/* Enable non-blocking I/O */
		flags = fcntl(fd, F_GETFL);
		if (fcntl(fd, F_SETFL, flags | O_NONBLOCK)) {
			perror("failed to set O_NONBLOCK");
			return 1;
		}
	}

	if (read(fd, &old_state, sizeof(old_state)) < 0) {
		perror("failed to read initial pin state");
		return 1;
	}
	new_state = old_state;
	printf("0x%08lx\n", old_state);

	if (use_async) {
		sigset_t sigio_mask, saved_mask;

		sigemptyset(&sigio_mask);
		sigaddset(&sigio_mask, SIGIO);

		/*
		 * In order to avoid race conditions with
		 * sigio_handler, we need to mask SIGIO while checking
		 * for updates to the pin state. sigsuspend() will
		 * atomically restore the sigmask (i.e. unmask SIGIO)
		 * and suspend the process until a signal comes along.
		 */
		sigprocmask(SIG_BLOCK, &sigio_mask, &saved_mask);
		while (1) {
			if (new_state != old_state) {
				printf("0x%08lx\n", new_state);
				old_state = new_state;
			}
			/*
			 * If SIGIO comes along here, we would miss an
			 * update.  However, that won't happen since
			 * we've masked it.
			 */
			sigsuspend(&saved_mask);

			/*
			 * sigsuspend() always returns with an
			 * error. If everything is normal, errno will
			 * be set to EINTR because we received a
			 * signal, and "Interrupted system call" will
			 * be displayed here.
			 */
			perror("sigsuspend");
		}
	} else {
		struct pollfd fds = {
			.fd = fd,
			.events = POLLIN,
		};

		while (1) {
			if (use_poll) {
				int ret;

				ret = poll(&fds, 1, 1000);
				if (ret < 0) {
					perror("poll");
					return 1;
				} else if (ret == 0) {
					/*
					 * This is normal if no
					 * buttons are pushed for one
					 * second.
					 */
					fprintf(stderr, "poll() timed out\n");
					continue;
				} else {
					/* ret should always be 1 here */
					fprintf(stderr, "poll() returned %d\n",
						ret);
				}
			}
			if (read(fd, &new_state, sizeof(new_state)) < 0) {
				/*
				 * If -n is given and not -p or -a,
				 * read() will fail with EAGAIN, and
				 * "Resource temporarily unavailable"
				 * will be printed regularly. This is
				 * normal.
				 *
				 * However, if -p is given, no errors
				 * should ever happen here, even with
				 * -n, as poll() will ensure that we
				 * never get here unless data is
				 * available for reading.
				 */
				perror("read error");
				sleep(1);
			} else {
				printf("0x%08lx\n", new_state);
				old_state = new_state;
			}
		}
	}
}
